import numpy as np

from pyvest.investment import Profit


class InvestmentState:
    def __init__(self, investment, quantity_by_ticker=None,
                 total_cost_by_ticker=None, average_cost_by_ticker=None,
                 total_commission_by_ticker=None, asset_price_by_ticker=None,
                 cash=None, total_invested=None, profit=None,
                 investment_state_to_copy=None):

        self.__investment = investment

        self.__init_values_by_ticker()

        self.__cash = 0
        self.__total_invested = 0
        self.__profit = Profit(self.__investment)

        if investment_state_to_copy is not None:
            self.__copy(investment_state_to_copy)

        self.__initialize_attributes(quantity_by_ticker, total_cost_by_ticker,
                                     average_cost_by_ticker,
                                     total_commission_by_ticker,
                                     asset_price_by_ticker, cash,
                                     total_invested, profit)

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    @property
    def quantity_by_ticker(self):
        return self.__quantity_by_ticker

    @quantity_by_ticker.setter
    def quantity_by_ticker(self, value):
        self.__quantity_by_ticker = value

    @property
    def total_cost_by_ticker(self):
        return self.__total_cost_by_ticker

    @total_cost_by_ticker.setter
    def total_cost_by_ticker(self, value):
        self.__total_cost_by_ticker = value

    @property
    def average_cost_by_ticker(self):
        return self.__average_cost_by_ticker

    @average_cost_by_ticker.setter
    def average_cost_by_ticker(self, value):
        self.__average_cost_by_ticker = value

    @property
    def total_commission_by_ticker(self):
        return self.__total_commission_by_ticker

    @total_commission_by_ticker.setter
    def total_commission_by_ticker(self, value):
        self.__total_commission_by_ticker = value

    @property
    def asset_price_by_ticker(self):
        return self.__asset_price_by_ticker

    @asset_price_by_ticker.setter
    def asset_price_by_ticker(self, value):
        self.__asset_price_by_ticker = value

    @property
    def value_by_ticker(self):
        return self.__calculate_value_by_ticker()

    @property
    def weight_by_ticker(self):
        return self.__calculate_weight_by_ticker()

    @property
    def cash(self):
        return self.__cash

    @cash.setter
    def cash(self, value):
        self.__cash = value

    @property
    def total_invested(self):
        return self.__total_invested

    @total_invested.setter
    def total_invested(self, value):
        self.__total_invested = value

    @property
    def profit(self):
        return self.__profit

    @profit.setter
    def profit(self, value):
        self.__profit = value

    def __init_values_by_ticker(self):

        self.__total_commission_by_ticker = {}
        self.__average_cost_by_ticker = {}
        self.__total_cost_by_ticker = {}
        self.__quantity_by_ticker = {}
        self.__asset_price_by_ticker = {}

        for ticker in self.__investment.assets.keys():
            self.__total_commission_by_ticker[ticker] = 0
            self.__average_cost_by_ticker[ticker] = 0
            self.__total_cost_by_ticker[ticker] = 0
            self.__quantity_by_ticker[ticker] = 0
            self.__asset_price_by_ticker[ticker] = 0

    def __copy(self, investment_state_to_copy):
        self.__quantity_by_ticker = \
            investment_state_to_copy.quantity_by_ticker.copy()
        self.__total_cost_by_ticker = \
            investment_state_to_copy.total_cost_by_ticker.copy()
        self.__average_cost_by_ticker = \
            investment_state_to_copy.average_cost_by_ticker.copy()
        self.__total_commission_by_ticker = \
            investment_state_to_copy.total_commission_by_ticker.copy()
        self.__asset_price_by_ticker = \
            investment_state_to_copy.asset_price_by_ticker.copy()
        self.__cash = investment_state_to_copy.cash
        self.__total_invested = investment_state_to_copy.total_invested

        self.__profit = Profit(self.__investment,
                               profit_to_copy=investment_state_to_copy.profit)

    def __initialize_attributes(self, quantity_by_ticker, total_cost_by_ticker,
                                average_cost_by_ticker,
                                total_commission_by_ticker,
                                asset_price_by_ticker, cash, total_invested,
                                profit):
        if quantity_by_ticker is not None:
            for ticker, _ in quantity_by_ticker.items():
                self.__quantity_by_ticker[ticker] = quantity_by_ticker[ticker]

        if total_cost_by_ticker is not None:
            for ticker, _ in total_cost_by_ticker.items():
                self.__total_cost_by_ticker[ticker] = \
                    total_cost_by_ticker[ticker]

        if average_cost_by_ticker is not None:
            for ticker, _ in average_cost_by_ticker.items():
                self.__average_cost_by_ticker[ticker] = \
                    average_cost_by_ticker[ticker]

        if total_commission_by_ticker is not None:
            for ticker, _ in total_commission_by_ticker.items():
                self.__total_commission_by_ticker[ticker] = \
                    total_commission_by_ticker[ticker]

        if asset_price_by_ticker is not None:
            for ticker, _ in asset_price_by_ticker.items():
                self.__asset_price_by_ticker[ticker] = asset_price_by_ticker[
                    ticker]

        if cash is not None:
            self.__cash = cash

        if total_invested is not None:
            self.__total_invested = total_invested

        if profit is not None:
            self.__profit = profit

    def __calculate_value_by_ticker(self):
        value_by_ticker = {}
        for ticker, quantity in self.__quantity_by_ticker.items():
            value_by_ticker[ticker] = \
                quantity * self.__asset_price_by_ticker[ticker]

        return value_by_ticker

    def __calculate_weight_by_ticker(self):
        weight_by_ticker = {}
        value_by_ticker = self.__calculate_value_by_ticker()
        total_value = sum(value_by_ticker.values())
        for ticker, quantity in self.__quantity_by_ticker.items():
            value = quantity * self.__asset_price_by_ticker[ticker]
            weight_by_ticker[ticker] = value / total_value if total_value > 0 \
                else np.NaN

        return weight_by_ticker

    def __generate_output(self):
        output = ""
        for ticker in self.__quantity_by_ticker.keys():
            output += ticker + "\n" \
                      + "  Quantity: " \
                      + str(self.__quantity_by_ticker[ticker]) + "\n" \
                      + "  Value: " \
                      + str(self.value_by_ticker[ticker]) + "\n" \
                      + "  Weight: " \
                      + str(self.weight_by_ticker[ticker]) + "\n" \
                      + "  Total Cost: " \
                      + str(self.__total_cost_by_ticker[ticker]) + "\n" \
                      + "  Average Cost: " \
                      + str(self.__average_cost_by_ticker[ticker]) + "\n" \
                      + "  Total Commission: " \
                      + str(self.__total_commission_by_ticker[ticker]) + "\n"
        output += "Cash: " + str(self.__cash) + "\n" \
                  + "Total Invested: " + str(self.__total_invested)

        if self.__profit is not None:
            output += "\n" + "Profit: \n" + str(self.__profit) + "\n" \
                      + "-------------------------------------\n"

        return output



import pyvest

import matplotlib.pyplot as plt
import numpy as np


class FactorModelVisualizer:

    def __init__(self, factor_models, labels=None, colors=None,
                 error_bars_colors=None):

        if isinstance(factor_models, pyvest.FactorModel):
            self.__factor_models = [factor_models]
        else:
            self.__factor_models = factor_models

        self.__labels = labels if labels is not None else ["1", "2", "3", "4"]
        self.__colors = colors if colors is not None \
            else ['blue', 'red', 'green', 'yellow']
        self.__error_bars_colors = error_bars_colors \
            if error_bars_colors is not None else ['C0', 'C1', 'C2', 'C3']

        self.__fig = None
        self.__ax = None
        self.__fig_sml = None
        self.__ax_sml = None

    ##################### fig ###################
    @property
    def fig(self):
        return self.__fig

    ##################### ax ###################
    @property
    def ax(self):
        return self.__ax

    ##################### fig_sml ###################
    @property
    def fig_sml(self):
        return self.__fig_sml

    ##################### ax_sml ###################
    @property
    def ax_sml(self):
        return self.__ax_sml

    ########################## PUBLIC ##########################

    def plot_realized_vs_predicted_average_return(self, min_return=0,
                                                  max_return=1.5,
                                                  legend='upper left',
                                                  portfolios=None,
                                                  confidence_level=None):

        self.__perform_factor_models_calculations(confidence_level)

        RETURN_STEP = 0.01

        # Set plot parameters
        self.__fig, self.__ax = plt.subplots(figsize=(16, 10))

        self.__ax.set_title("Realized vs. predicted average return",
                            fontsize=30)
        self.__ax.set_xlabel("Predicted average return", fontsize=30)
        self.__ax.set_ylabel("Realized average return", fontsize=30)
        self.__ax.set_xticks(np.arange(0, 2, step=0.2))
        self.__ax.tick_params(axis='both', labelsize=25)

        # Plot predicted vs realized average returns
        predicted_average_return_line_array = np.arange(min_return, max_return,
                                                        RETURN_STEP)
        self.__ax.plot(predicted_average_return_line_array,
                       predicted_average_return_line_array, color='black',
                       linewidth=2)

        labels_iter = iter(self.__labels)
        colors_iter = iter(self.__colors)
        error_bars_colors_iter = iter(self.__error_bars_colors)
        for factor_model in self.__factor_models:

            if isinstance(portfolios, dict):
                shown_portfolios = portfolios[factor_model.name]
            else:
                shown_portfolios = portfolios if portfolios is not None \
                    else list(factor_model.Y.columns)

            label = next(labels_iter)
            color = next(colors_iter)
            error_bars_color = next(error_bars_colors_iter)

            lower_error_bars = [rr.lower_error_bar for ptf, rr
                                in factor_model.items() if ptf
                                in shown_portfolios]
            upper_error_bars = [rr.upper_error_bar for ptf, rr
                                in factor_model.items() if ptf
                                in shown_portfolios]
            predicted_average_returns = \
                [ret for ptf, ret in
                 factor_model.predicted_average_returns.items()
                 if ptf in shown_portfolios]
            realized_average_returns = \
                [ret for ptf, ret in
                 factor_model.realized_average_returns.items()
                 if ptf in shown_portfolios]

            self.__ax.errorbar(predicted_average_returns,
                               realized_average_returns,
                               linestyle="None",
                               marker='.',
                               markersize=15,
                               yerr=[lower_error_bars, upper_error_bars],
                               capsize=5,
                               color=error_bars_color,
                               markeredgecolor=color,
                               markerfacecolor=color,
                               linewidth=1,
                               label=label)

            # For loop to annotate all points
            for ptf in shown_portfolios:
                self.__ax.annotate(ptf,
                                   (factor_model.predicted_average_returns[ptf],
                                    factor_model.realized_average_returns[
                                        ptf] + 0.02),
                                   fontsize=20)
        if type(legend) is str:
            self.__ax.legend(fontsize=15, loc=legend)

    def plot_sml(self, beta_min=0, beta_max=2, legend='upper left',
                 portfolios=None, confidence_level=None):

        self.__perform_factor_models_calculations(confidence_level)

        self.__check_if_one_factor()
        self.__check_factor_models_use_same_r_f()
        self.__check_factor_models_use_same_factors()

        beta_array, sml_array = self.__calculate_sml(
            self.__factor_models[0].r_f,
            self.__factor_models[0].X.drop('const', axis=1).squeeze(),
            beta_min, beta_max)

        # Set up plot parameters
        self.__fig_sml, self.__ax_sml = plt.subplots(figsize=(16, 10))

        self.__ax_sml.set_title("SML and average return against " + r"$\beta$",
                                fontsize=30)
        self.__ax_sml.set_xlabel(r"$\hat{\beta}$", fontsize=30)
        self.__ax_sml.set_ylabel("Realized average return", fontsize=30)
        self.__ax_sml.set_xticks(np.arange(0, 2, step=0.2))
        self.__ax_sml.tick_params(axis='both', labelsize=25)

        # Plot the SML
        self.__ax_sml.plot(beta_array, sml_array, color='black', linewidth=2)

        labels_iter = iter(self.__labels)
        colors_iter = iter(self.__colors)
        error_bars_colors_iter = iter(self.__error_bars_colors)
        for factor_model in self.__factor_models:

            if isinstance(portfolios, dict):
                shown_portfolios = portfolios[factor_model.name]
            else:
                shown_portfolios = portfolios if portfolios is not None \
                    else list(factor_model.Y.columns)

            label = next(labels_iter)
            color = next(colors_iter)
            error_bars_color = next(error_bars_colors_iter)

            lower_error_bars = [rr.lower_error_bar for ptf, rr
                                in factor_model.items() if ptf
                                in shown_portfolios]
            upper_error_bars = [rr.upper_error_bar for ptf, rr
                                in factor_model.items() if ptf
                                in shown_portfolios]
            realized_average_returns = \
                [ret for ptf, ret in
                 factor_model.realized_average_returns.items()
                 if ptf in shown_portfolios]

            beta_list = [rr.beta for ptf, rr in factor_model.items() if ptf
                         in shown_portfolios]

            self.__ax_sml.errorbar(beta_list,
                                   realized_average_returns,
                                   linestyle="None",
                                   marker='.',
                                   markersize=15,
                                   yerr=[lower_error_bars, upper_error_bars],
                                   capsize=5,
                                   color=error_bars_color,
                                   markeredgecolor=color,
                                   markerfacecolor=color,
                                   linewidth=1,
                                   label=label)

            # For loop to annotate all points
            for ptf in shown_portfolios:
                self.__ax_sml.annotate(ptf,
                                       (factor_model[ptf].beta,
                                        factor_model.realized_average_returns[
                                            ptf] + 0.02),
                                       fontsize=20)
        if type(legend) is str:
            self.__ax_sml.legend(fontsize=15, loc=legend)

    ########################## PRIVATE ##########################

    def __check_factor_models_use_same_factors(self):
        X = None
        for factor_model in self.__factor_models:
            if X is None:
                X = factor_model.X
            else:
                if not factor_model.X.equals(X):
                    raise ValueError(
                        "The factor models must use the same factors.")

    def __check_factor_models_use_same_r_f(self):
        r_f = None
        for factor_model in self.__factor_models:
            if r_f is None:
                r_f = factor_model.r_f
            else:
                if not factor_model.r_f.equals(r_f):
                    raise ValueError(
                        "The factor models must use the same r_f.")

    def __check_if_one_factor(self):
        for factor_model in self.__factor_models:
            if len(factor_model.X.columns) != 2:
                raise ValueError(
                    "One of the factor models uses more than one factor.")

    def __calculate_sml(self, r_f, mkt_rf_series, beta_min=0, beta_max=2):

        BETA_STEP = 0.01

        rf_mean = r_f.mean()
        mkt_rf_mean = mkt_rf_series.mean()

        beta_array = np.arange(beta_min, beta_max, BETA_STEP)
        sml_array = rf_mean + mkt_rf_mean * beta_array

        return beta_array, sml_array

    def __perform_factor_models_calculations(self, confidence_level):
        for factor_model in self.__factor_models:
            if factor_model.realized_average_returns is None:
                factor_model.calculate_realized_vs_predicted_average_returns(
                    return_results=False)
            for regression in factor_model.values():
                regression.calculate_error_bar(
                    confidence_level=confidence_level, return_results=True)



import math
import numpy as np

from pyvest.general.general import calculate_portfolio_expected_return
from pyvest.general.general import calculate_portfolio_standard_deviation


class Portfolio:

    def __init__(self, weights, mu, cov, r_f=None, expense_ratios=None,
                 expense_ratio_r_f=None, assets=None, nb_decimal_places=4):

        self.__mu = np.array(mu)
        self.__cov = np.array(cov)
        self.__r_f = r_f
        self.__assign_weights(weights, r_f)
        self.__assign_expense_ratios(expense_ratios, expense_ratio_r_f)
        self.__assets = assets
        self.__nb_decimal_places = nb_decimal_places

        if self.__assets is not None:
            self.__assets_weights = {asset: weight for (asset, weight)
                                     in zip(self.__assets, self.__weights)}
            if r_f is not None:
                self.__assets_weights["r_f"] = self.__weights[-1]
        else:
            self.__assets_weights = list(self.__weights)

    def __repr__(self):
        return self.__generate_output()

    def __str__(self):
        return self.__generate_output()

    ##################### mu ###################
    @property
    def mu(self):
        return self.__mu

    @mu.setter
    def mu(self, value):
        self.__mu = value
        # TODO: Test whether value is an 'array'.
        # if type(value) is float or type(value) is int:
        #     self.__mu = value
        # else:
        #     raise TypeError(
        #         "The attribute 'mu' must be of type float or int.")

    ##################### cov ###################
    @property
    def cov(self):
        return self.__cov

    @cov.setter
    def cov(self, value):
        self.__cov = value

    ##################### r_f ###################
    @property
    def r_f(self):
        return self.__r_f

    @r_f.setter
    def r_f(self, value):
        if type(value) is float or type(value) is int:
            self.__r_f = value
        else:
            raise TypeError(
                "The attribute 'r_f' must be of type float or int.")

    ##################### weights ###################
    @property
    def weights(self):
        return self.__weights

    @weights.setter
    def weights(self, value):
        self.__assign_weights(value, self.__r_f)

    ##################### expense_ratios ###################
    @property
    def expense_ratios(self):
        return self.__expense_ratios

    @expense_ratios.setter
    def expense_ratios(self, value):
        self.__assign_expense_ratios(value, self.__expense_ratio_r_f)

    ##################### expense_ratio_r_f ###################
    @property
    def expense_ratio_r_f(self):
        return self.__expense_ratio_r_f

    @expense_ratio_r_f.setter
    def expense_ratio_r_f(self, value):
        self.__assign_expense_ratios(self.__expense_ratios, value)

    ##################### nb_decimal_places ###################
    @property
    def nb_decimal_places(self):
        return self.__nb_decimal_places

    @nb_decimal_places.setter
    def nb_decimal_places(self, value):
        self.__nb_decimal_places = value

    ##################### read-only attributes ###################
    @property
    def expected_return(self):
        effective_mu = self.__calculate_effective_mu(
            self.__augmented_mu, self.__expense_ratios,
            self.__expense_ratio_r_f)
        return calculate_portfolio_expected_return(self.__augmented_weights,
                                                   effective_mu)

    @property
    def standard_deviation(self):
        return calculate_portfolio_standard_deviation(self.__augmented_weights,
                                                      self.__augmented_cov)

    @property
    def augmented_weights(self):
        return self.__augmented_weights

    @property
    def augmented_mu(self):
        return self.__augmented_mu

    @property
    def augmented_cov(self):
        return self.__augmented_cov

    ########################## PRIVATE ##########################

    def __assign_weights(self, weights, r_f):
        if type(weights) is not list and type(weights) is not np.ndarray:
            raise TypeError("The attribute 'weights' has to be a list or a "
                            "np.array.")
        if len(weights) != len(self.__mu) and r_f is None:
            raise ValueError("The length of 'weights' must be equal to {}"
                             .format(len(self.__mu)))
        if len(weights) != len(self.__mu) + 1 and r_f is not None:
            raise ValueError("The length of 'weights' must be equal to {}"
                             .format(len(self.__mu) + 1))
        if not math.isclose(sum(weights), 1.0):
            raise ValueError("The sum of the weights must be equal to 1. "
                             "sum(weights)={}".format(sum(weights)))

        self.__weights = np.array(weights)

        if self.__r_f is not None:
            self.__augmented_mu = np.concatenate((self.__mu,
                                                  [self.__r_f]))
            self.__augmented_weights = self.__weights

            zeros_column = (np.zeros((len(self.__cov), 1)))
            zeros_row = (np.zeros((1, len(self.__cov) + 1)))
            self.__augmented_cov = \
                np.concatenate(
                    (np.concatenate((self.__cov, zeros_column), axis=1),
                     zeros_row))
        else:
            # self.__augmented_mu = np.concatenate((self.__mu, [0.0]))
            # self.__augmented_weights = np.concatenate((self.__weights, [0.0]))
            self.__augmented_mu = self.__mu
            self.__augmented_weights = self.__weights
            self.__augmented_cov = self.__cov

        return self.__weights

    def __assign_expense_ratios(self, expense_ratios, expense_ratio_r_f):

        if expense_ratio_r_f is None:
            self.__expense_ratio_r_f = expense_ratio_r_f
        elif type(expense_ratio_r_f) is int or type(expense_ratio_r_f) \
                is float:
            self.__expense_ratio_r_f = expense_ratio_r_f
        else:
            raise TypeError("The attribute 'expense_ratio_r_f' has to be of "
                            "type 'int' or 'float'.")

        if expense_ratios is None:
            self.__expense_ratios = expense_ratios
        elif type(expense_ratios) is int or type(expense_ratios) is float:
            self.__expense_ratios = np.repeat(expense_ratios,
                                              len(self.__mu))
        elif type(expense_ratios) is not list and type(expense_ratios) \
                is not np.array:
            raise TypeError("The attribute 'weights' has to be of type 'int', "
                            "'float', 'list' or 'np.array'.")
        elif len(expense_ratios) != len(self.__mu):
            raise ValueError("The length of 'expense_ratios' must be equal to"
                             " {}".format(len(self.__mu)))
        else:
            self.__expense_ratios = np.array(expense_ratios)

        return self.__expense_ratios, self.__expense_ratio_r_f

    def __calculate_effective_mu(self, augmented_mu, expense_ratios,
                                 expense_ratio_r_f):

        effective_mu = augmented_mu.copy()
        if expense_ratios is not None:
            effective_mu[:-1] = \
                (1 + augmented_mu[:-1]) * (1 - expense_ratios) - 1
        if expense_ratio_r_f is not None:
            effective_mu[-1] = \
                (1 + augmented_mu[-1]) * (1 - expense_ratio_r_f) - 1

        return effective_mu

    def __generate_output(self):
        if type(self.__assets_weights) is list:
            rounded_assets_weights = [round(weight, self.__nb_decimal_places)
                                      for weight in self.__assets_weights]
        else:
            rounded_assets_weights = \
                {asset: round(weight, self.__nb_decimal_places)
                 for asset, weight in self.__assets_weights.items()}

        rounded_expected_return = round(self.expected_return,
                                        self.__nb_decimal_places)

        rounded_standard_deviation = round(self.standard_deviation,
                                           self.__nb_decimal_places)

        output = "weights: {}\nexpected return: {}\nstandard deviation: " \
                 "{}".format(str(rounded_assets_weights),
                             str(rounded_expected_return),
                             str(rounded_standard_deviation))

        return output



import statsmodels.api as sm


class Regression:
    def __init__(self, X, Y, name=None):

        self.__X = X
        self.__Y = Y
        self.__name = name

        self.__regression_results = None
        self.__alpha = None
        self.__beta = None
        self.__conf_int_df = None
        self.__lower_error_bar = None
        self.__upper_error_bar = None

        # Default confidence level
        self.__confidence_level = 0.95

    def __str__(self):
        ret = self.__name if self.__name is not None else ""
        if self.regression_results is not None:
            ret = self.regression_results.summary().__str__()

        return ret

    def __repr__(self):
        ret = self.__name if self.__name is not None else ""
        if self.regression_results is not None:
            ret = self.regression_results.summary().__repr__()

        return ret

    def _repr_html_(self):
        ret = self.__name if self.__name is not None else ""
        if self.regression_results is not None:
            ret = self.regression_results.summary().as_html()

        return ret

    def _repr_latex_(self):
        ret = self.__name if self.__name is not None else ""
        if self.regression_results is not None:
            ret = self.regression_results.summary().as_latex()

        return ret

    @property
    def X(self):
        return self.__X

    @property
    def Y(self):
        return self.__Y

    @property
    def name(self):
        return self.__name

    @property
    def regression_results(self):
        return self.__regression_results

    @property
    def alpha(self):
        return self.__alpha

    @property
    def beta(self):
        return self.__beta

    @property
    def confidence_interval(self):
        if self.__conf_int_df is None:
            self.calculate_confidence_interval(return_results=False)

        return self.__conf_int_df

    @property
    def error_bar(self):
        if self.__lower_error_bar is None or self.__upper_error_bar is None:
            self.calculate_error_bar(return_results=False)

        return self.__lower_error_bar, self.__upper_error_bar

    @property
    def lower_error_bar(self):
        if self.__lower_error_bar is None:
            self.calculate_error_bar(return_results=False)

        return self.__lower_error_bar

    @property
    def upper_error_bar(self):
        if self.__upper_error_bar is None:
            self.calculate_error_bar(return_results=False)

        return self.__upper_error_bar

    @property
    def confidence_level(self):
        return self.__confidence_level

    def calculate_regression(self, return_results=True):
        linear_model = sm.OLS(self.__Y, self.__X)
        self.__regression_results = linear_model.fit()

        estimated_parameters = self.__regression_results.params
        self.__alpha = estimated_parameters[0]
        self.__beta = estimated_parameters[1:]

        if return_results:
            return self.__regression_results

    def calculate_confidence_interval(self, confidence_level=None,
                                      return_results=True):

        if confidence_level is not None:
            self.__confidence_level = confidence_level

        self.__conf_int_df = self.__regression_results.conf_int(
            1 - self.__confidence_level)
        self.__conf_int_df.rename({0: 'Lower bound', 1: 'Upper bound'}, axis=1,
                                  inplace=True)

        if return_results:
            return self.__conf_int_df

    def calculate_error_bar(self, confidence_level=None, return_results=True):

        conf_int_df = self.calculate_confidence_interval(confidence_level)
        self.__upper_error_bar = \
            conf_int_df.loc['const']['Upper bound'] - self.__alpha
        self.__lower_error_bar = self.__alpha - conf_int_df.loc['const'][
            'Lower bound']

        if return_results:
            return self.__lower_error_bar, self.__upper_error_bar



import numpy as np

import math

from pyvest.general.portfolio import Portfolio
from pyvest.general.general import calculate_portfolio_standard_deviation
from pyvest.general.general import calculate_portfolio_sharpe_ratio
from pyvest.investment_universe.investment_universe_visualizer import \
    InvestmentUniverseVisualizer

from scipy.optimize import minimize
from scipy.optimize import LinearConstraint
from scipy.optimize import Bounds

from pyvest.investment_universe.investor import Investor
from pyvest.investment_universe.portfolios_generator import \
    FeasiblePortfoliosGeneratorRandom, FeasiblePortfoliosGeneratorFrontier


class InvestmentUniverse:
    MAX_NB_ITERATIONS = 100

    def __init__(self, assets, mu, cov, r_f=None, min_weight=0,
                 min_weight_r_f=None, parameters=None):

        self.__assign_parameters(parameters)

        self.__assets = assets

        self.__nb_risky_assets = len(mu)
        self.__nb_assets = self.__nb_risky_assets if r_f is None \
            else self.__nb_risky_assets + 1

        self.__assign_r_f(r_f)
        self.__assign_mu(mu)
        self.__assign_cov(cov)
        self.__assign_min_weight(min_weight)
        self.__assign_min_weight_r_f(min_weight_r_f)

        self.__investors = {}
        self.__nb_unnamed_investors = 0

        self.__calculate_assets_std()

        self.__feasible_portfolios = None
        self.__feasible_portfolios_with_r_f = None
        self.__feasible_portfolios_surface = None
        self.__feasible_portfolios_surface_with_r_f = None
        self.__feasible_portfolios_equation = None
        self.__feasible_portfolios_equation_with_r_f = None
        self.__mvp = None
        self.__efficient_frontier = None
        self.__efficient_frontier_equation = None
        self.__tangency_portfolio = None
        self.__cal = None
        self.__cal_equation = None
        self.__other_portfolios = None
        self.__market_portfolio = None
        self.__total_wealth = None
        self.__investors = {}

        self.__min_weights_bound = None
        self.__sum_weights_assets_equals_one_constraint = None

        self.__efficient_mu_min = None
        self.__efficient_mu_max = None

        self.__visualizer = None

        self.__feasible_portfolios_generator_random = None
        self.__feasible_portfolios_generator_frontier = None
        self.__feasible_portfolios_generator_random_r_f = None
        self.__feasible_portfolios_generator_frontier_r_f = None

    ################################# ATTRIBUTES ##############################

    @property
    def assets(self):
        return self.__assets

    @assets.setter
    def assets(self, value):
        self.__assets = value

    @property
    def mu(self):
        return self.__mu

    @mu.setter
    def mu(self, value):
        self.__mu = np.array(value)

    @property
    def augmented_mu(self):
        return self.__augmented_mu

    @augmented_mu.setter
    def augmented_mu(self, value):
        self.__augmented_mu = np.array(value)

    @property
    def cov(self):
        return self.__cov

    @cov.setter
    def cov(self, value):
        self.__cov = np.array(value)

    @property
    def augmented_cov(self):
        return self.__augmented_cov

    @augmented_cov.setter
    def augmented_cov(self, value):
        self.__augmented_cov = np.array(value)

    @property
    def r_f(self):
        return self.__r_f

    @r_f.setter
    def r_f(self, value):
        self.__assign_r_f(value)

    @property
    def min_weight(self):
        return self.__min_weights

    @min_weight.setter
    def min_weight(self, value):
        self.__assign_min_weight(value)

    @property
    def min_weight_r_f(self):
        return self.__min_weight_r_f

    @min_weight_r_f.setter
    def min_weight_r_f(self, value):
        self.__assign_min_weight_r_f(value)

    @property
    def parameters(self):
        return self.__parameters

    @parameters.setter
    def parameters(self, value):
        self.__assign_parameters(value)

    @property
    def std(self):
        return self.__std

    @property
    def feasible_portfolios(self):
        return self.__feasible_portfolios

    @property
    def feasible_portfolios_with_r_f(self):
        return self.__feasible_portfolios_with_r_f

    @property
    def feasible_portfolios_surface(self):
        return self.__feasible_portfolios_surface

    @property
    def feasible_portfolios_surface_with_r_f(self):
        return self.__feasible_portfolios_surface_with_r_f

    @property
    def feasible_portfolios_equation(self):
        return self.__feasible_portfolios_equation

    @property
    def feasible_portfolios_equation_with_r_f(self):
        return self.__feasible_portfolios_equation_with_r_f

    @property
    def mvp(self):
        return self.__mvp

    @property
    def efficient_frontier(self):
        return self.__efficient_frontier

    @property
    def efficient_frontier_equation(self):
        return self.__efficient_frontier_equation

    @property
    def tangency_portfolio(self):
        return self.__tangency_portfolio

    @property
    def cal(self):
        return self.__cal

    @property
    def cal_equation(self):
        return self.__cal_equation

    @property
    def other_portfolios(self):
        return self.__other_portfolios

    @property
    def visualizer(self):
        return self.__visualizer

    @property
    def investors(self):
        return self.__investors

    @property
    def market_portfolio(self):
        return self.__market_portfolio

    @property
    def total_wealth(self):
        return self.__total_wealth

    ################################# PUBLIC FUNCTIONS #################################

    def calculate_feasible_portfolios(self, nb_portfolios=20000,
                                      with_r_f=False,
                                      random=False,
                                      frontier=True):

        if self.__min_weights is None and not with_r_f:
            self.__feasible_portfolios_equation = \
                self.__calculate_feasible_portfolio_equation
        elif self.__min_weights is None and with_r_f:
            tangency_portfolio = self.__calculate_tangency_portfolio() \
                if self.tangency_portfolio is None else self.tangency_portfolio
            self.__feasible_portfolios_equation_with_r_f = \
                lambda x: self.__calculate_cal_equation(x, tangency_portfolio)
        else:
            self.__calculate_feasible_portfolios_with_min_weights(
                nb_portfolios, with_r_f, random)

    def __calculate_feasible_portfolios_with_min_weights(
            self, nb_portfolios=20000, with_r_f=False, random=False):

        if random and with_r_f:
            if self.__feasible_portfolios_generator_random_r_f is None:
                self.__feasible_portfolios_generator_random_r_f = \
                    FeasiblePortfoliosGeneratorRandom(self, with_r_f)
            portfolios_generator = \
                self.__feasible_portfolios_generator_random_r_f
        elif random and not with_r_f:
            if self.__feasible_portfolios_generator_random is None:
                self.__feasible_portfolios_generator_random = \
                    FeasiblePortfoliosGeneratorRandom(self, with_r_f)
            portfolios_generator = self.__feasible_portfolios_generator_random
        elif with_r_f:
            if self.__feasible_portfolios_generator_frontier_r_f is None:
                self.__feasible_portfolios_generator_frontier_r_f = \
                    FeasiblePortfoliosGeneratorFrontier(self, with_r_f)
            portfolios_generator = \
                self.__feasible_portfolios_generator_frontier_r_f
        else:
            if self.__feasible_portfolios_generator_frontier is None:
                self.__feasible_portfolios_generator_frontier = \
                    FeasiblePortfoliosGeneratorFrontier(self, with_r_f)
            portfolios_generator = \
                self.__feasible_portfolios_generator_frontier

        portfolios_generator.generate_portfolios(
            nb_portfolios=nb_portfolios, distance=0.05)
        if with_r_f:
            self.__feasible_portfolios_surface_with_r_f = \
                portfolios_generator.frontier
        else:
            self.__feasible_portfolios_surface = portfolios_generator.frontier

    def calculate_mvp(self, x0=None):

        # Initial guess (seed value)
        if x0 is None:
            x0 = np.ones(self.__nb_risky_assets) / self.__nb_risky_assets

        # Assign results
        self.__mvp = self.__calculate_mvp(x0)

        return self.__mvp

    def calculate_efficient_portfolio(self, expected_return=None,
                                      standard_deviation=None, name=None,
                                      x0=None, tolerance=None):

        tolerance = self.__parameters["optimization_tolerance"] \
            if tolerance is None else tolerance

        # Initial guess (seed value)
        if x0 is None:
            x0 = np.ones(self.__nb_risky_assets) / self.__nb_risky_assets

        mvp = self.__calculate_mvp(x0) if not self.__mvp else self.__mvp
        self.__calculate_efficient_mu_min_max(mvp)

        if expected_return is not None and standard_deviation is not None:
            raise ValueError("Only one of 'mu' and 'sigma' must be passed as "
                             "argument.")
        elif expected_return is not None:
            efficient_portfolio = \
                self.__calculate_efficient_portfolio_from_mu(expected_return,
                                                             x0, tolerance)
        elif standard_deviation is not None:
            efficient_portfolio = \
                self.__calculate_efficient_portfolio_from_sigma(
                    standard_deviation, x0, tolerance)
        else:
            raise ValueError("Either 'mu' or 'sigma' must be passed as "
                             "argument.")

        self.calculate_portfolio(efficient_portfolio, name)

        return efficient_portfolio

    def calculate_efficient_frontier(self, nb_portfolios=1000,
                                     x0=None,
                                     tolerance=None,
                                     return_portfolios=False):
        if self.__min_weights is None:
            mvp_mu_sigma = self.__calculate_mvp_mu_sigma_from_equation()
            self.__efficient_frontier_equation = {
                "equation": self.__calculate_feasible_portfolio_equation,
                "mvp": mvp_mu_sigma
            }
            efficient_frontier = self.__efficient_frontier_equation
        else:
            efficient_frontier = \
                self.__calculate_efficient_frontier_with_min_weights(
                    nb_portfolios, x0, tolerance, return_portfolios)

        if return_portfolios:
            return efficient_frontier

    def __calculate_efficient_frontier_with_min_weights(
            self, nb_portfolios=1000, x0=None, tolerance=None,
            return_portfolios=False):

        tolerance = self.__parameters["optimization_tolerance"] \
            if tolerance is None else tolerance

        # Initial guess (seed value)
        if x0 is None:
            x0 = np.ones(self.__nb_risky_assets) / self.__nb_risky_assets

        mvp = self.__calculate_mvp(x0) if not self.__mvp else self.__mvp

        self.__calculate_efficient_mu_min_max(mvp)
        efficient_mu_array = self.__calculate_efficient_mu_array(nb_portfolios)

        # Calculate the efficient portfolios
        self.__efficient_frontier = []
        for efficient_mu in efficient_mu_array:
            efficient_portfolio = \
                self.__calculate_efficient_portfolio_from_mu(efficient_mu, x0,
                                                             tolerance)
            self.__efficient_frontier.append(efficient_portfolio)

        if return_portfolios:
            return self.__efficient_frontier

    def calculate_tangency_portfolio(self, x0=None, tolerance=None):

        self.__tangency_portfolio = \
            self.__calculate_tangency_portfolio(x0, tolerance)

        return self.__tangency_portfolio

    def __calculate_tangency_portfolio(self, x0=None, tolerance=None):

        if self.__r_f is None:
            raise ValueError("You need to add a risk-free asset first!")

        tolerance = self.__parameters["optimization_tolerance"] \
            if tolerance is None else tolerance

        # Initial guess (seed value)
        if x0 is None:
            x0 = np.ones(self.__nb_risky_assets) / self.__nb_risky_assets

        # Sum portfolio weights equals 1 constraint
        self.__sum_weights_assets_equals_one_constraint = LinearConstraint(
            np.ones(self.__nb_risky_assets), 1, 1)

        tangency_portfolio_result = minimize(
            lambda x: -calculate_portfolio_sharpe_ratio(x, self.__mu,
                                                        self.__cov,
                                                        self.__r_f),
            x0,
            bounds=self.__min_weights_bound,
            constraints=[self.__sum_weights_assets_equals_one_constraint],
            tol=tolerance)

        tangency_portfolio_weights = np.append(tangency_portfolio_result.x,
                                               [0])

        tangency_portfolio = Portfolio(tangency_portfolio_weights, self.__mu,
                                       self.__cov, r_f=self.__r_f,
                                       assets=self.__assets)

        return tangency_portfolio

    def calculate_cal(self, return_portfolios=False):

        tangency_portfolio = self.__calculate_tangency_portfolio() \
            if self.tangency_portfolio is None else self.tangency_portfolio

        if self.__min_weights is None:
            self.__cal_equation = \
                lambda x: self.__calculate_cal_equation(x, tangency_portfolio)
        else:
            self.__calculate_cal_with_min_weights(tangency_portfolio)

        if return_portfolios:
            return self.__cal

    def __calculate_cal_with_min_weights(self, tangency_portfolio):

        min_fraction, max_fraction, step_fraction = self.__get_cal_parameters()

        self.__cal = []
        for tangency_weight in np.arange(min_fraction,
                                         max_fraction,
                                         step_fraction):
            cal_weights = tangency_weight * tangency_portfolio.weights
            cal_weights[-1] = 1 - tangency_weight

            cal_portfolio = Portfolio(cal_weights, self.__mu,
                                      self.__cov, self.__r_f,
                                      assets=self.__assets)
            self.__cal.append(cal_portfolio)

        return self.__cal

    def calculate_portfolio_on_cal(self, expected_return=None,
                                   standard_deviation=None,
                                   weight_r_f=None, weight_tangency=None,
                                   name=None):

        nb_not_none = sum([
            x is not None for x in [expected_return, standard_deviation,
                                    weight_r_f, weight_tangency]])
        if nb_not_none != 1:
            raise ValueError("One and only one of 'mu', 'sigma', 'weight_r_f' "
                             "and 'weight_tangency' must be passed as "
                             "argument.")

        tangency_portfolio = self.__calculate_tangency_portfolio() \
            if self.tangency_portfolio is None else self.tangency_portfolio

        if expected_return is not None:
            standard_deviation = self.__calculate_cal_equation(
                expected_return, tangency_portfolio)

        if standard_deviation is not None:
            weight_tangency = \
                standard_deviation / tangency_portfolio.standard_deviation

        if weight_r_f is not None:
            weight_tangency = 1 - weight_r_f

        if self.__min_weights is not None:
            _, max_fraction, _ = self.__get_cal_parameters()
            if weight_tangency > max_fraction:
                raise ValueError("The value of 'weight_tangency' must be less "
                                 "than {}.".format(max_fraction))

        portfolio_weights = weight_tangency * tangency_portfolio.weights
        portfolio_weights[-1] = 1 - weight_tangency

        portfolio = self.calculate_portfolio(portfolio_weights, name)

        return portfolio

    def calculate_portfolio(self, portfolio, name=None):

        if isinstance(portfolio, Portfolio):
            portfolio_obj = portfolio
        elif (isinstance(portfolio, list)
              or isinstance(portfolio, np.ndarray)
              or isinstance(portfolio, tuple)) \
                and ((len(portfolio) == self.__nb_risky_assets
                      and self.__r_f is None)
                     or (len(portfolio) == self.__nb_risky_assets + 1
                         and self.__r_f is not None)):
            portfolio_obj = Portfolio(portfolio, self.__mu, self.__cov,
                                      r_f=self.__r_f, assets=self.__assets)
        else:
            raise TypeError("The variable 'portfolio' must be an object of "
                            "type Portfolio or a list of weights of dimension "
                            "{}.".format(self.__nb_assets))

        if self.__other_portfolios is None:
            self.__other_portfolios = {}

        self.__other_portfolios[tuple(portfolio_obj.weights)] = \
            (portfolio_obj, name)

        return portfolio_obj

    def remove_portfolio(self, portfolio=None):
        if portfolio is None:
            self.__other_portfolios = None
        elif isinstance(portfolio, list) or isinstance(portfolio, tuple) \
                or isinstance(portfolio, np.ndarray):
            del self.__other_portfolios[tuple(portfolio)]
        elif isinstance(portfolio, Portfolio):
            del self.__other_portfolios[portfolio.weights]
        elif isinstance(portfolio, str):
            keys_to_delete = []
            for key, value in self.__other_portfolios.items():
                if value[1] == portfolio:
                    keys_to_delete.append(key)
            for key in keys_to_delete:
                del self.__other_portfolios[key]
        else:
            raise TypeError("The parameter 'portfolio' must be either a "
                            "Portfolio, a list of weights, or a string!")

    def get_portfolio_by_name(self, portfolio_name):
        portfolio = None
        if self.__other_portfolios is not None:
            for weights, (ptf_obj, name) in self.__other_portfolios.items():
                if name == portfolio_name:
                    portfolio = ptf_obj

        return portfolio

    def plot(self, compare_with=None, labels=None, weights_visible=True,
             zoom_individual=False, min_expected_return=None,
             max_expected_return=None, min_standard_deviation=None,
             max_standard_deviation=None, investors=None,
             indifference_curves=None, optimal_portfolios=True,
             legend='upper left'):
        investment_universes = [self]
        if isinstance(compare_with, InvestmentUniverse):
            investment_universes.append(compare_with)
        elif isinstance(compare_with, list):
            investment_universes.extend(compare_with)

        if self.__visualizer is None:
            self.__visualizer = InvestmentUniverseVisualizer(
                investment_universes, labels=labels,
                weights_visible=weights_visible)
        else:
            self.__visualizer.investment_universes = investment_universes
            self.__visualizer.labels = labels
            self.__visualizer.reset_colors()
        self.__visualizer.plot(zoom_individual=zoom_individual,
                               min_expected_return=min_expected_return,
                               max_expected_return=max_expected_return,
                               min_standard_deviation=min_standard_deviation,
                               max_standard_deviation=max_standard_deviation,
                               investors=investors,
                               indifference_curves=indifference_curves,
                               optimal_portfolios=optimal_portfolios,
                               legend=legend)

    def add_investor(self, gamma, wealth=None, portfolio=None,
                     utility_function=None, name=None):

        if name is None:
            self.__nb_unnamed_investors += 1
            name = "Investor {}".format(self.__nb_unnamed_investors)

        investor = Investor(self, gamma, wealth, portfolio, utility_function,
                            name)

        self.__investors[name] = investor

        return investor

    def calculate_market_portfolio(self):

        nb_weights = self.__nb_risky_assets + 1 if self.__r_f is not None \
            else self.__nb_risky_assets

        market_assets_value = np.zeros(nb_weights)
        self.__total_wealth = 0
        for investor in self.__investors.values():
            self.__total_wealth += investor.wealth
            weights = investor.portfolio.weights
            market_assets_value += np.array(weights) * investor.wealth

        market_weights = market_assets_value / self.__total_wealth

        if self.__r_f is not None:
            market_weights = \
                market_weights / sum(market_weights[:self.__nb_risky_assets])
            market_weights[-1] = 0.0

        self.__market_portfolio = Portfolio(market_weights, self.__mu,
                                            self.__cov, r_f=self.__r_f,
                                            assets=self.__assets)

        return self.__market_portfolio

    def calculate_optimal_portfolios(self, investors=None):

        if investors is None:
            investors = self.investors
        elif type(investors) is str:
            investors = [investors]

        optimal_portfolios = {}
        for investor_name in investors:
            investor = self.investors[investor_name]
            optimal_portfolio = investor.calculate_optimal_portfolio()
            optimal_portfolios[investor_name] \
                = {
                "portfolio": optimal_portfolio,
                "utility": investor.optimal_portfolio_utility
            }

        return optimal_portfolios

    ########################## PRIVATE ##########################

    def __assign_mu(self, mu):
        self.mu = mu
        if self.__r_f is not None:
            self.__augmented_mu = np.concatenate((self.__mu,
                                                  [self.__r_f]))
        else:
            self.__augmented_mu = self.mu

    def __assign_cov(self, cov):
        self.cov = cov
        if self.__r_f is not None:
            zeros_column = (np.zeros((len(self.__cov), 1)))
            zeros_row = (np.zeros((1, len(self.__cov) + 1)))
            self.__augmented_cov = \
                np.concatenate(
                    (np.concatenate((self.__cov, zeros_column), axis=1),
                     zeros_row))
        else:
            self.__augmented_cov = self.cov

    def __assign_r_f(self, r_f):
        if r_f is None or type(r_f) is float or type(r_f) is int:
            self.__r_f = r_f
        else:
            raise TypeError(
                "The parameter 'r_f' must be of type float or int.")

    def __assign_min_weight(self, min_weight):
        if min_weight is None:
            self.__min_weights = None
        elif type(min_weight) is float or type(min_weight) is int:
            self.__min_weights = min_weight * np.ones(self.__nb_risky_assets)
        elif type(min_weight) is list or type(min_weight) is np.array:
            self.__min_weights = min_weight
        else:
            raise TypeError(
                "The parameter 'min_weight' must be of type float ,int or "
                "list, or None.")

    def __assign_min_weight_r_f(self, min_weight_r_f):
        if min_weight_r_f is None and self.__min_weights is None:
            self.__min_weight_r_f = None
        elif min_weight_r_f is None:
            self.__min_weight_r_f = self.__parameters["min_weight_r_f"]
        else:
            self.__min_weight_r_f = min_weight_r_f

    def __assign_parameters(self, parameters):
        if parameters is None:
            self.__assign_default_parameters()
        else:
            self.__check_parameters(parameters)
            self.__parameters = parameters

    def __assign_default_parameters(self):
        self.__parameters = {
            "optimization_tolerance": 1e-8,
            "cal_min_fraction": 0,
            "cal_step_fraction": 0.001,
            "min_weight_r_f": -4,
            "cal_max_std": 100
        }

    def __check_parameters(self, parameters):
        if type(parameters) is not dict:
            raise TypeError(
                "The variable 'parameters' must be a dictionary.")

        if "optimization_tolerance" not in parameters:
            raise ValueError(
                "The dictionary parameters must contain the key "
                "'optimization_tolerance'.")
        if type(parameters["optimization_tolerance"]) is not float:
            raise TypeError(
                "The value of 'optimization_tolerance' must be of type float.")

    def __calculate_assets_std(self):
        std = []
        for i in range(0, len(self.__assets)):
            std.append(math.sqrt(self.__cov[i][i]))
        self.__std = np.array(std)
        return self.__std

    def __calculate_mvp(self, x0):

        # Sum portfolio weights equals 1 constraint
        self.__sum_weights_assets_equals_one_constraint = LinearConstraint(
            np.ones(self.__nb_risky_assets), 1, 1)

        self.__min_weights_bound = Bounds(self.__min_weights, np.inf) \
            if self.__min_weights is not None else None

        # Minimize
        mvp_result = minimize(
            lambda x: calculate_portfolio_standard_deviation(x, self.__cov),
            x0,
            bounds=self.__min_weights_bound,
            constraints=[self.__sum_weights_assets_equals_one_constraint],
            tol=self.__parameters["optimization_tolerance"])

        if self.__r_f is not None:
            mvp_weights = \
                np.append(mvp_result.x, [0])
        else:
            mvp_weights = mvp_result.x

        return Portfolio(mvp_weights, self.__mu, self.__cov, r_f=self.__r_f,
                         assets=self.__assets)

    def __calculate_efficient_mu_min_max(self, mvp):
        self.__efficient_mu_min = mvp.expected_return

        if self.__min_weights is not None:
            mu_argmax = np.argmax(self.__mu)
            mu_max = self.__mu[mu_argmax]
            others_mu = np.delete(self.__mu, mu_argmax)
            others_min_weight = np.delete(self.__min_weights,
                                          mu_argmax)
            self.__efficient_mu_max = (1 - sum(
                others_min_weight)) * mu_max + np.dot(
                others_mu, others_min_weight)
        else:
            LARGE_FACTOR = 2
            self.__efficient_mu_max = LARGE_FACTOR * max(self.mu)

    def __calculate_efficient_mu_array(self, nb_portfolios):
        # Define the range of expected return over which to calculate the
        # efficient frontier.
        delta_mu = self.__efficient_mu_max - self.__efficient_mu_min
        step = delta_mu / nb_portfolios
        efficient_mu_array = np.arange(self.__efficient_mu_min,
                                       self.__efficient_mu_max, step)

        return efficient_mu_array

    def __calculate_efficient_portfolio_from_mu(self, mu, x0, tolerance,
                                                maximize=False):

        factor = -1 if maximize else 1

        efficient_mu_constraint = LinearConstraint(self.__mu.T, mu, mu)
        efficient_portfolio_result = minimize(
            lambda x: factor * calculate_portfolio_standard_deviation(
                x, self.__cov),
            x0,
            bounds=self.__min_weights_bound,
            constraints=[self.__sum_weights_assets_equals_one_constraint,
                         efficient_mu_constraint],
            tol=tolerance)
        if not efficient_portfolio_result.success:
            raise ValueError(
                "minimize was not successful with bounds={} and constraints={}!"
                .format(self.__min_weights_bound, mu))

        if self.__r_f is not None:
            efficient_portfolio_weights = \
                np.append(efficient_portfolio_result.x, [0])
        else:
            efficient_portfolio_weights = efficient_portfolio_result.x

        efficient_portfolio = Portfolio(efficient_portfolio_weights, self.__mu,
                                        self.__cov, r_f=self.__r_f,
                                        assets=self.__assets)

        return efficient_portfolio

    def __calculate_efficient_portfolio_from_sigma(self, sigma, x0, tolerance):

        nb_iter = 0

        mu_min = self.__efficient_mu_min
        mu_max = self.__efficient_mu_max
        tentative_mu = (mu_min + mu_max) / 2
        tentative_portfolio = \
            self.__calculate_efficient_portfolio_from_mu(tentative_mu, x0,
                                                         tolerance)
        tentative_sigma = tentative_portfolio.standard_deviation
        while abs(tentative_sigma - sigma) >= tolerance:
            if sigma - tentative_sigma > 0:
                mu_min = tentative_mu
                tentative_mu = (tentative_mu + mu_max) / 2
            else:
                mu_max = tentative_mu
                tentative_mu = (mu_min + tentative_mu) / 2
            tentative_portfolio = \
                self.__calculate_efficient_portfolio_from_mu(tentative_mu, x0,
                                                             tolerance)
            tentative_sigma = tentative_portfolio.standard_deviation
            nb_iter += 1
            if nb_iter > self.MAX_NB_ITERATIONS:
                raise StopIteration("Number of iterations exceeded "
                                    "MAX_NB_ITERATIONS ({})"
                                    .format(self.MAX_NB_ITERATIONS))

        return tentative_portfolio

    def __get_cal_parameters(self):

        min_fraction = self.__parameters["cal_min_fraction"]
        step_fraction = self.__parameters["cal_step_fraction"]

        if self.__efficient_frontier is not None:
            max_std = max([ptf.standard_deviation
                           for ptf in self.__efficient_frontier])
        else:
            max_std = self.__parameters["cal_max_std"]

        if self.__tangency_portfolio is not None:
            max_fraction = min(
                1.0 - self.min_weight_r_f,
                max_std / self.__tangency_portfolio.standard_deviation)
        else:
            max_fraction = 1.0 - self.min_weight_r_f

        return min_fraction, max_fraction, step_fraction

    def __calculate_feasible_portfolio_equation(self, mu):

        one_vector = np.ones((len(self.mu), 1))

        s_1_1 = float(np.dot(one_vector.T,
                             np.dot(np.linalg.inv(self.cov), one_vector)))
        s_1_mu = float(np.dot(one_vector.T,
                              np.dot(np.linalg.inv(self.cov), self.mu)))
        s_mu_mu = float(
            np.dot(self.mu.T, np.dot(np.linalg.inv(self.cov), self.mu)))
        d = s_mu_mu * s_1_1 - s_1_mu ** 2

        sigma = math.sqrt(1 / d * (
                s_1_1 * mu ** 2 - 2 * s_1_mu * mu + s_mu_mu))

        return sigma

    def __calculate_cal_equation(self, mu, tangency_portfolio):

        tang_ptf_std = tangency_portfolio.standard_deviation
        tang_ptf_exp_ret = tangency_portfolio.expected_return
        slope = tang_ptf_std / (tang_ptf_exp_ret - self.r_f)

        sigma = slope * abs(mu - self.r_f)

        return sigma

    def __calculate_mvp_mu_sigma_from_equation(self):

        one_vector = np.ones((len(self.mu), 1))

        s_1_1 = float(np.dot(one_vector.T,
                             np.dot(np.linalg.inv(self.cov), one_vector)))
        s_1_mu = float(np.dot(one_vector.T,
                              np.dot(np.linalg.inv(self.cov), self.mu)))

        mvp_mu = s_1_mu / s_1_1
        mvp_sigma = self.__calculate_feasible_portfolio_equation(mvp_mu)

        return mvp_mu, mvp_sigma



from pyvest.data_reader.returns_data_reader import *
from pyvest.data_reader.factors_data_reader import *
from pyvest.data_reader.y_f_data_reader import *
from pyvest.data_reader.fama_french_data_reader import *


__all__ = ['ReturnsDataReader', 'YFDataReader', 'FactorsDataReader',
           'FamaFrenchDataReader']



import math
import itertools

import numpy as np

from scipy.optimize import LinearConstraint, minimize, Bounds

import pyvest.general as general


class PortfoliosGenerator:

    def generate_portfolios(self, nb_portfolios, distance):
        raise NotImplementedError(
            "PortfoliosGenerator.generate_portfolios has not been "
            "implemented.")


class FeasiblePortfoliosGenerator(PortfoliosGenerator):

    def __init__(self, investment_universe, with_r_f=False):
        self.__investment_universe = investment_universe
        self.__with_r_f = with_r_f

    def generate_portfolios(self, nb_portfolios, distance):
        raise NotImplementedError(
            "FeasiblePortfoliosGenerator.generate_portfolios has not been "
            "implemented.")

    @property
    def investment_universe(self):
        return self.__investment_universe

    @property
    def with_r_f(self):
        return self.__with_r_f


class FeasiblePortfoliosGeneratorRandom(FeasiblePortfoliosGenerator):

    def __init__(self, investment_universe, with_r_f=False):
        super().__init__(investment_universe, with_r_f)

        self.__mu = investment_universe.mu
        self.__cov = investment_universe.cov
        self.__r_f = investment_universe.r_f
        self.__assets = investment_universe.assets

        if with_r_f is False:
            self.__min_weight = investment_universe.min_weight
        else:
            self.__min_weight = np.append(investment_universe.min_weight,
                                          investment_universe.min_weight_r_f)

    def generate_portfolios(self, nb_portfolios, distance=None):

        feasible_portfolios = []
        for i in range(1, nb_portfolios):
            risky_assets_portfolio_weights = \
                self.__calculate_random_portfolio_weights()

            if not self.with_r_f and self.__r_f is not None:
                portfolio_weights = \
                    np.append(risky_assets_portfolio_weights, [0])
            else:
                portfolio_weights = risky_assets_portfolio_weights

            portfolio = general.Portfolio(portfolio_weights, self.__mu,
                                          self.__cov, r_f=self.__r_f,
                                          assets=self.__assets)
            feasible_portfolios.append(portfolio)

        return feasible_portfolios

    def __calculate_random_portfolio_weights(self):
        # This function adds random portfolio weights
        # The argument "smallest_weight" denotes the smallest weight
        # admissible for a given asset. For example, "smallest_weight=0"
        # indicates that short sales are not allowed, and "smallest_weight=-1"
        # implies that the weight of each asset in the portfolio must be equal
        # or greater to -1. The function returns an array of portfolio weights

        weights = np.random.dirichlet(np.ones(len(self.__min_weight)),
                                      size=1)[0]
        norm_weights = \
            weights * (1 - sum(self.__min_weight)) + self.__min_weight

        return norm_weights


class FeasiblePortfoliosGeneratorFrontier(FeasiblePortfoliosGenerator):

    def __init__(self, investment_universe, with_r_f=False):
        super().__init__(investment_universe, with_r_f)

        if with_r_f:
            self.__augmented_mu = investment_universe.augmented_mu
            self.__augmented_cov = investment_universe.augmented_cov

        self.__mu = investment_universe.mu
        self.__cov = investment_universe.cov

        self.__r_f = investment_universe.r_f
        self.__assets = investment_universe.assets

        self.__nb_risky_assets = len(self.__mu)
        self.__nb_assets = self.__nb_risky_assets if not with_r_f \
            else self.__nb_risky_assets + 1

        if not with_r_f:
            self.__min_weight = investment_universe.min_weight
        else:
            self.__min_weight = np.append(investment_universe.min_weight,
                                          investment_universe.min_weight_r_f)

        self.__portfolios_list = None
        self.__frontier = None
        self.__min_portfolios_list = None
        self.__max_portfolios_list = None
        self.__filtered_max_portfolios_list = None
        self.__filtered_min_portfolios_list = None

        self.__step = None
        self.__factor = None

        self.__max_portfolios_list = []

    @property
    def frontier(self):
        return self.__frontier

    def generate_portfolios(self, nb_portfolios=None, distance=None):

        self.__portfolios_list = []
        self.__frontier = []
        self.__min_portfolios_list = []

        x0 = np.ones(self.__nb_assets) / self.__nb_assets

        tolerance = 1e-8

        self.__calculate_max_portfolios()
        self.__filter_max_portfolios()

        self.__portfolios_list = []
        for ptf_max in self.__filtered_max_portfolios_list:

            mu = ptf_max.expected_return
            ptf_min = self.__calculate_frontier_portfolio_from_mu(
                mu, x0, tolerance, maximize=False)

            if ptf_min is not None:
                self.__min_portfolios_list.append(ptf_min)
                self.__portfolios_list.append(ptf_min)
                self.__portfolios_list.append(ptf_max)
                self.__frontier.append((ptf_min, ptf_max))

        return self.__portfolios_list

    def __calculate_frontier_portfolio_from_mu(self, mu, x0, tolerance,
                                               maximize=False):

        efficient_portfolio = None

        factor = -1 if maximize else 1

        self.__min_weights_bound = Bounds(self.__min_weight,
                                          np.inf)

        if self.with_r_f:
            # Sum portfolio weights equals 1 constraint
            self.__sum_weights_assets_equals_one_constraint = LinearConstraint(
                np.ones(self.__nb_assets), 1, 1)

            efficient_mu_constraint = LinearConstraint(self.__augmented_mu.T,
                                                       mu, mu)

            efficient_portfolio_result = minimize(
                lambda
                    x: factor * general.calculate_portfolio_standard_deviation(
                    x, self.__augmented_cov),
                x0,
                bounds=self.__min_weights_bound,
                constraints=[self.__sum_weights_assets_equals_one_constraint,
                             efficient_mu_constraint],
                tol=tolerance)
        else:
            # Sum portfolio weights equals 1 constraint
            self.__sum_weights_assets_equals_one_constraint = LinearConstraint(
                np.ones(self.__nb_assets), 1, 1)

            efficient_mu_constraint = LinearConstraint(self.__mu.T, mu, mu)
            efficient_portfolio_result = minimize(
                lambda
                    x: factor * general.calculate_portfolio_standard_deviation(
                    x, self.__cov),
                x0,
                bounds=self.__min_weights_bound,
                constraints=[self.__sum_weights_assets_equals_one_constraint,
                             efficient_mu_constraint],
                tol=tolerance)

        if not efficient_portfolio_result.success:
            print("minimize was not successful with bounds={} and "
                  "constraints={}!".format(self.__min_weights_bound, mu))
        else:

            if self.__r_f is not None and not self.with_r_f:
                efficient_portfolio_weights = \
                    np.append(efficient_portfolio_result.x, [0])
            else:
                efficient_portfolio_weights = efficient_portfolio_result.x

            efficient_portfolio = general.Portfolio(
                efficient_portfolio_weights,
                self.__mu, self.__cov,
                r_f=self.__r_f,
                assets=self.__assets)

        return efficient_portfolio

    def __calculate_max_portfolios(self):

        step = 0.0001

        index_pairs_list = itertools.combinations(range(self.__nb_assets), 2)

        for index_pair in index_pairs_list:
            weights_array = np.zeros(self.__nb_assets)
            for x in np.arange(0, 1 + step, step):
                weights_array[index_pair[0]] = x
                weights_array[index_pair[1]] = 1 - x
                portfolio = self.__create_portfolio(weights_array)
                self.__max_portfolios_list.append(portfolio)

    def __filter_max_portfolios(self):

        mu_step = 0.005

        highest_std_ptfs_dict = {}
        for ptf in self.__max_portfolios_list:
            mu_index = math.floor(ptf.expected_return / mu_step)
            if mu_index not in highest_std_ptfs_dict \
                    or (ptf.standard_deviation
                        > highest_std_ptfs_dict[mu_index].standard_deviation):
                highest_std_ptfs_dict[mu_index] = ptf

        self.__filtered_max_portfolios_list = highest_std_ptfs_dict.values()

    def __create_portfolio(self, normalized_weights):

        portfolio_weights = self.__inverse_normalize_weights(
            normalized_weights)

        if self.__r_f is not None and not self.with_r_f:
            portfolio_weights = \
                np.append(portfolio_weights, [0])

        portfolio = general.Portfolio(portfolio_weights,
                                      self.__mu,
                                      self.__cov, r_f=self.__r_f,
                                      assets=self.__assets)

        return portfolio

    def __inverse_normalize_weights(self, normalized_weights):

        weights = normalized_weights * (1 - sum(self.__min_weight)) \
                  + self.__min_weight

        return weights

    def __normalize_weights(self, weights):

        normalized_weights = (weights - self.__min_weight) \
                             / (1 - sum(self.__min_weight))

        return normalized_weights



from pyvest.factor_model.factor_model import FactorModel


class CAPM(FactorModel):
    def __init__(self, factors, returns, r_f_column="RF",
                 mkt_rf_column="Mkt-RF", name="CAPM"):

        r_f = factors[r_f_column]
        capm_factors = factors[mkt_rf_column]

        super().__init__(r_f, capm_factors, returns, name)

    def plot(self, compare_with=None, labels=None, colors=None,
             error_bars_colors=None, legend='upper left', min_return=0,
             max_return=1.5, portfolios=None, confidence_level=None, sml=True,
             beta_min=0, beta_max=2):
        super().plot(compare_with=compare_with, labels=labels, colors=colors,
                     error_bars_colors=error_bars_colors, legend=legend,
                     min_return=min_return, max_return=max_return,
                     portfolios=portfolios, confidence_level=confidence_level,
                     sml=sml, beta_min=beta_min, beta_max=beta_max)



from pyvest.factor_model.factor_model import FactorModel


class FF3F(FactorModel):
    def __init__(self, factors, returns, r_f_column="RF",
                 mkt_rf_column="Mkt-RF", smb_column="SMB", hml_column="HML",
                 name="FF3F"):

        r_f = factors[r_f_column]
        ff3f_factors = factors[[mkt_rf_column, smb_column, hml_column]]

        super().__init__(r_f, ff3f_factors, returns, name)



from setuptools import setup

setup(
    package_dir={'pyvest': 'pyvest'},
    packages=["pyvest", "pyvest.data_reader", "pyvest.factor_model",
              "pyvest.general", "pyvest.investment",
              "pyvest.investment_universe", "pyvest.simulation"]
)



from pyvest.investment.investment import *
from pyvest.investment.asset import *
from pyvest.investment.transaction import *
from pyvest.investment.investment_visualizer import *
from pyvest.investment.investment_state import *
from pyvest.investment.investment_dated_info import *

__all__ = ['Investment', 'InvestmentState', 'InvestmentVisualizer', 'Stock',
           'Profit', 'InvestmentDatedInfo']



from pyvest.data_reader import ReturnsDataReader
from pyvest.data_reader.history_data_reader import HistoryDataReader

import pandas as pd
import yfinance as yf


class YFDataReader(ReturnsDataReader, HistoryDataReader):
    def __init__(self):
        super().__init__()

    def read_returns(self, tickers, start_date, end_date, freq='1M'):

        if isinstance(tickers, str):
            tickers = [tickers]

        monthly_returns_df = pd.DataFrame()
        for ticker in tickers:
            monthly_returns_df[ticker] = self.__read_ticker_returns(
                ticker, start_date, end_date, freq)

        return monthly_returns_df

    def read_history(self, ticker, start_date, end_date, interval='1d'):

        history_df = yf.Ticker(ticker).history(start=start_date, end=end_date,
                                               interval=interval,
                                               auto_adjust=False)

        return history_df

    def __read_ticker_returns(self, ticker, start_date, end_date, freq):
        history_df = yf.Ticker(ticker).history(start=start_date, end=end_date,
                                               auto_adjust=False)
        monthly_history_df = history_df.resample(freq).last()
        monthly_returns_series = monthly_history_df[
                                     'Adj Close'].pct_change() * 100
        monthly_returns_series = monthly_returns_series[1:]

        return monthly_returns_series



from pyvest.investment_universe import *
from pyvest.general import *
from pyvest.data_reader import *
from pyvest.factor_model import *
from pyvest.investment import *

__all__ = ['InvestmentUniverse', 'InvestmentUniverseVisualizer', 'Portfolio',
           'ReturnsDataReader', 'YFDataReader', 'Investor', 'FactorModel',
           'FactorModelVisualizer', 'FactorsDataReader',
           'FamaFrenchDataReader', 'CAPM', 'FF3F', 'Investment', 'Stock']
